"""
The module with CLI code that handles the `hopla version` command.
"""

import logging

import click

log = logging.getLogger()

MAJOR_VERSION = 0
MINOR_VERSION = 0
PATCH_VERSION = 12
PRE_RELEASE = "alpha"


def hopla_version() -> str:
    """
    Prints the hopla version (Not to be confused with the habitica API version)

    The output should satisfy [semantic versioning](https://semver.org/)
    TODO: verify if this is the case
    """

    version_core = f"{MAJOR_VERSION}.{MINOR_VERSION}.{PATCH_VERSION}"
    if PRE_RELEASE is not None and PRE_RELEASE != "":
        pre_release = f"-{PRE_RELEASE}"
    else:
        pre_release = ""

    return version_core + pre_release


@click.command()
def version():
    """Print the Hopla version string.

    \f
    :return:
    """
    log.debug("hopla version")
    click.echo(hopla_version())



#!/usr/bin/env python3
from hopla.cli.groupcmds.get_user import HabiticaUser


class TestHabiticaUser:

    def test_get_stats(self):
        user_test_stat_values = {
            "buffs": {
                "str": 50, "int": 50, "per": 3206, "con": 50, "stealth": 0, "streaks": False,
                "snowball": False, "spookySparkles": False, "shinySeed": False, "seafoam": False
            },
            "training": {"int": 0, "per": 0, "str": 0, "con": 0},
            "hp": 50, "mp": 65.7, "exp": 2501, "gp": 1072327.9, "lvl": 121,
            "class": "wizard", "points": 0, "str": 0, "con": 0, "int": 12, "per": 88,
            "toNextLevel": 5010, "maxHealth": 50, "maxMP": 304
        }

        user = HabiticaUser(user_dict={"stats": user_test_stat_values})

        assert user.get_stats() == user_test_stat_values

    def test_get_auth(self):
        user_test_auth_values = {
            "local": {"username": "hopla", "lowerCaseUsername": "hopla",
                      "email": "something+habitica@gmail.com"
                      },
            "timestamps": {"created": "2022-03-22T24:23:38.119Z",
                           "loggedin": "2022-09-18T08:47:45.286Z",
                           "updated": "2022-09-18T14:20:55.530Z"
                           },
            "facebook": {}, "google": {}, "apple": {}
        }

        user = HabiticaUser(user_dict={"auth": user_test_auth_values})

        assert user.get_auth() == user_test_auth_values



from hopla.cli.groupcmds.buy import times_until_poor


class TestTimesUntilPoor:
    def test_times_until_poor__budget_is0(self):
        assert times_until_poor(0.) == 0

    def test_times_until_poor__budget_is_below100(self):
        assert times_until_poor(92.3) == 0

    def test_times_until_poor__budget_is_over100(self):
        assert times_until_poor(212) == 2



#!/usr/bin/env python3
import random

import pytest

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.zoo.fooddata import FoodData
from hopla.hoplalib.zoo.foodmodels import FeedingStatus, InvalidFeedingStatus
from hopla.hoplalib.zoo.petdata import PetData
from hopla.hoplalib.zoo.petmodels import InvalidPet, \
    Pet, PetMountPair, Zoo, ZooBuilder, ZooHelper
from tests.testutils.user_test_utils import UserTestUtil

_SAMPLE_SIZE = 50
"""
Sample size to use for pytest.mark.parameterize when unittests become slow
"""


class TestPetInit:
    def test_init_raises_invalid_pet_name(self):
        with pytest.raises(InvalidPet) as execinfo:
            invalid_pet_name = "INVALID_PET"
            Pet(invalid_pet_name)

        err_msg = str(execinfo.value)
        assert invalid_pet_name in err_msg
        assert f"an issue at {GlobalConstants.NEW_ISSUE_URL}" in err_msg

    @pytest.mark.parametrize(
        "pet_name",
        # Testing all pets takes too long (approx. 2 seconds), so we sample.
        random.sample(PetData.pet_names, k=_SAMPLE_SIZE)
    )
    def test_init_valid_pet_name_ok(self, pet_name: str):
        pet = Pet(pet_name=pet_name)
        assert pet.pet_name == pet_name

    @pytest.mark.parametrize(
        "invalid_feed_status",
        [-10, -2, 1, 3, 51, 100]
    )
    def test_init_raises_invalid_feeding_status_fail(self, invalid_feed_status: int):
        with pytest.raises(InvalidFeedingStatus) as execinfo:
            feeding_status = FeedingStatus(invalid_feed_status)
            Pet("PandaCub-IcySnow", feeding_status=feeding_status)

        err_msg = str(execinfo.value)
        assert f"feeding_status={int(invalid_feed_status)}" in err_msg

    @pytest.mark.parametrize(
        "valid_feed_status",
        [5, 10, 20, -1, 49]
    )
    def test_init_sets_valid_feeding_status_ok(self, valid_feed_status: int):
        feeding_status = FeedingStatus(valid_feed_status)
        pet = Pet("LionCub-Sunset", feeding_status=feeding_status)

        assert valid_feed_status == int(pet.feeding_status)


class TestFavoriteFood:
    @pytest.mark.parametrize(
        "pet_name,expected_favorite_food", [
            ("Wolf-Base", "Meat"),
            ("FlyingPig-CottonCandyPink", "CottonCandyPink"),
            ("PandaCub-CottonCandyBlue", "CottonCandyBlue"),
            ("TigerCub-Golden", "Honey"),
            ("Fox-Red", "Strawberry"),
            ("BearCub-White", "Milk"),
            ("Dragon-Desert", "Potatoe"),
            ("Robot-Shade", "Chocolate"),
            ("Dolphin-Skeleton", "Fish")
        ]
    )
    def test_favorite_food_for_pets_with_favorite_food_ok(self, pet_name: str,
                                                          expected_favorite_food: str):
        pet = Pet(pet_name)
        assert pet.favorite_food() == expected_favorite_food

    @pytest.mark.parametrize(
        "pet_name", random.sample(PetData.only_1favorite_food_pet_names, k=_SAMPLE_SIZE)
    )
    def test_favorite_food_single_favorite_food_ok(self, pet_name: str):
        pet = Pet(pet_name)
        alternative_food = "SomeFakeFood"

        assert "Any" != pet.favorite_food()
        assert alternative_food != pet.favorite_food(
            default_value_for_all_favorite_food=alternative_food
        )
        assert "Unfeedable" != pet.favorite_food(
            default_value_for_unfeedable=alternative_food
        )

    @pytest.mark.parametrize(
        # only magic potion pets like all foods
        "pet_name", random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE)
    )
    def test_favorite_food_multiple_favorite_food_ok(self, pet_name: str):
        pet = Pet(pet_name)

        alternative_food = "Whatever"
        default_value_for_all_favorite_food_ok_pets = "Any"
        assert alternative_food == pet.favorite_food(
            default_value_for_all_favorite_food=alternative_food
        )
        assert default_value_for_all_favorite_food_ok_pets == pet.favorite_food(
            default_value_for_unfeedable=alternative_food
        )
        assert default_value_for_all_favorite_food_ok_pets == pet.favorite_food()

    @pytest.mark.parametrize(
        # only magic potion pets like all foods
        "pet_name", PetData.unfeedable_pet_names
    )
    def test_favorite_food_for_unfeedable_pets(self, pet_name: str):
        pet = Pet(pet_name)

        alternative_food = "Whatever"
        default_value_for_unfeedable_pets = "Unfeedable"
        assert default_value_for_unfeedable_pets == pet.favorite_food(
            default_value_for_all_favorite_food=alternative_food
        )
        assert alternative_food == pet.favorite_food(default_value_for_unfeedable=alternative_food)
        assert default_value_for_unfeedable_pets == pet.favorite_food()


class TestIsFavoriteFood:

    @pytest.mark.parametrize(
        "pet_name,food_name",
        # we use choices here because both groups are smaller than
        # the specified k
        zip(random.choices(PetData.unfeedable_pet_names, k=_SAMPLE_SIZE),
            random.choices(FoodData.drop_food_names, k=_SAMPLE_SIZE))
    )
    def test_is_favorite_food_false_because_unfeedable(self, pet_name: str,
                                                       food_name: str):
        pet = Pet(pet_name)
        assert pet.is_favorite_food(food_name) is False

    @pytest.mark.parametrize(
        "pet_name,food_name",
        zip(random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE),
            random.choices(FoodData.drop_food_names, k=_SAMPLE_SIZE))
    )
    def test_is_favorite_food_true_because_likes_all_food(self,
                                                          pet_name: str,
                                                          food_name: str):
        assert Pet(pet_name).is_favorite_food(food_name)

    @pytest.mark.parametrize(
        "pet_name,food_name,expected_result", [
            ("Fox-Red", "Strawberry", True),
            ("Fox-Red", "Milk", False),
            ("Parrot-Desert", "Potatoe", True),
            ("Parrot-Base", "Potatoe", False),
            ("Hippo-Golden", "Honey", True),
            ("Hippo-Golden", "CottonCandyBlue", False),
        ]
    )
    def test_is_favorite_food_for_hatching_pets(self,
                                                pet_name: str,
                                                food_name: str,
                                                expected_result: bool):
        pet = Pet(pet_name)
        result = pet.is_favorite_food(food_name)
        assert result is expected_result


class TestOtherPetFunctions:
    """ Test class for Pet functions that aren't covered elsewhere """
    PET_NAME = "Fox-Ember"

    def test___repr__(self):
        pet_name = "Fox-Shadow"
        feeding_status = FeedingStatus(10)
        pet = Pet(pet_name, feeding_status=feeding_status)

        result = str(pet)

        expected = f"Pet(self.pet_name='{pet_name}', self.feeding_status={feeding_status})"
        assert result == expected

    @pytest.mark.parametrize(
        "feeding_status,partial_expected_message", [
            (-1, f"can't feed self.pet_name='{PET_NAME}' you only have the mount"),
            (5, f"Cannot determine if self.pet_name='{PET_NAME}' can be fed"),
            (20, f"pet_name='{PET_NAME}' can be fed")
        ]
    )
    def test_feeding_status_explanation_ok(self, feeding_status: int,
                                           partial_expected_message: str):
        feeding_status = FeedingStatus(feeding_status)
        pet = Pet(TestOtherPetFunctions.PET_NAME, feeding_status=feeding_status)

        result_explanation = pet.feeding_status_explanation()

        assert partial_expected_message in result_explanation

    @pytest.mark.parametrize(
        "unfeedable_pet_name", [
            "Fox-Veteran", "JackOLantern-Glow", "JackOLantern-RoyalPurple", "MantisShrimp-Base",
            "Gryphon-Gryphatrice", "PandaCub-Dessert", "LionCub-Veggie", "LionCub-Dessert",
            "Fox-Veggie", "Fox-Dessert",
        ]
    )
    def test_feeding_status_unfeedable_pet(self, unfeedable_pet_name: str):
        pet = Pet(unfeedable_pet_name)

        result_explanation = pet.feeding_status_explanation()

        assert pet.pet_name in result_explanation
        assert "can't be fed because it is special" in result_explanation

    @pytest.mark.parametrize(
        "pet_name,is_gen1_expected", [
            ("Wolf-Base", True),
            ("FlyingPig-CottonCandyPink", True),
            ("PandaCub-CottonCandyBlue", True),
            ("TigerCub-Golden", True),
            ("Fox-Red", True),
            ("BearCub-White", True),
            ("Dragon-Desert", True),
            ("Cactus-Amber", False),
            ("BearCub-Frost", False),
            ("Butterfly-Red", False)
        ]
    )
    def test_is_generation1_pet_ok(self, pet_name, is_gen1_expected):
        pet = Pet(pet_name)
        assert pet.is_generation1_pet() is is_gen1_expected

    @pytest.mark.parametrize(
        "pet_name",
        random.sample(PetData.generation1_pet_names + PetData.quest_pet_names,
                      k=_SAMPLE_SIZE)
    )
    def test_has_just_1_favorite_food_true(self, pet_name):
        pet = Pet(pet_name=pet_name)
        assert pet.has_just_1_favorite_food()

    @pytest.mark.parametrize(
        "pet_name",
        random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE)
    )
    def test_has_just_1_favorite_food_false(self, pet_name):
        pet = Pet(pet_name=pet_name)
        assert pet.has_just_1_favorite_food() is False

    @pytest.mark.parametrize(
        "pet_name,likes_all_food_expected", [
            ("Fox-Golden", False),
            ("LionCub-Holly", True),
            ("PandaCub-Cupid", True),
            ("Octopus-Red", False),
            ("Seahorse-CottonCandyBlue", False)
        ]
    )
    def test_likes_all_food(self, pet_name: str,
                            likes_all_food_expected: bool):
        pet = Pet(pet_name)

        result = pet.likes_all_food()

        assert result is likes_all_food_expected

    @pytest.mark.parametrize(
        "pet_name,expected_from_drop_hatching_potion", [
            ("Fox-Golden", True),
            ("Dragon-Shade", True),
            ("LionCub-Holly", False),
            ("PandaCub-Cupid", False),
            ("Octopus-Red", True),
            ("Cheetah-CottonCandyPink", True),
            ("Seahorse-CottonCandyBlue", True),
            ("Unicorn-Skeleton", True),
            ("MagicalBee-Base", False),  # although -Base postfix, this is a world boss reward
            ("Phoenix-Base", False),  # although -Base postfix, this is a world boss reward
            ("Turkey-Base", False),  # although -Base postfix, this is a event sequence reward
            ("Gryphon-Gryphatrice", False),
        ]
    )
    def test_is_from_drop_hatching_potion(self, pet_name: str,
                                          expected_from_drop_hatching_potion: bool):
        pet = Pet(pet_name)

        result = pet.is_from_drop_hatching_potions()

        assert result is expected_from_drop_hatching_potion

    @pytest.mark.parametrize(
        "pet,food_name,expected_times", [
            (Pet("Hedgehog-Shade", feeding_status=FeedingStatus(10)), "Chocolate", 8),
            (Pet("Hedgehog-Shade", feeding_status=FeedingStatus(10)), "Milk", 20),
            (Pet("Egg-Base"), "Meat", 9),
            (Pet("Egg-Zombie"), "CottonCandyPink", 23),
            (Pet("Egg-Base", feeding_status=FeedingStatus(40)), "Meat", 2),
            (Pet("Egg-Base", feeding_status=FeedingStatus(44)), "Meat", 2),
            (Pet("Egg-Golden", feeding_status=FeedingStatus(47)), "Fish", 2),
        ]
    )
    def test_required_food_items_until_mount(self, pet: Pet,
                                             food_name: str,
                                             expected_times: int):
        times_result: int = pet.required_food_items_until_mount(food_name)

        assert times_result == expected_times


class TestPetMountPair:
    def test_feed_status_fail(self):
        pet = Pet("LionCub-Watery", feeding_status=None)

        with pytest.raises(InvalidFeedingStatus) as execinfo:
            PetMountPair(pet,
                         pet_available=True,
                         mount_available=False)

        err_msg = str(execinfo.value)
        assert f"PetMountPair requires that pet={pet} has a feeding status" in err_msg

    def test___repr__(self):
        feed_status = FeedingStatus(20)
        pet = Pet("Monkey-Skeleton", feeding_status=feed_status)
        pet_available = True
        mount_available = True

        pair = PetMountPair(pet,
                            pet_available=pet_available,
                            mount_available=mount_available)
        result = str(pair)

        expected = ("PetMountPair({"
                    "'pet': Pet(self.pet_name='Monkey-Skeleton', "
                    "self.feeding_status=FeedingStatus(20)), "
                    "'pet_available': True, ""'mount_available': True})")
        assert result == expected

    def test_can_feed_pet_mount_available(self):
        # pet itself is feedable, but mount is available,
        # so cant feed pet should return False
        pet = Pet("Wolf-Base", feeding_status=FeedingStatus(5))
        pair = PetMountPair(pet, pet_available=True, mount_available=True)

        result = pair.can_feed_pet()

        assert result is False

    def test_can_feed_pet_pet_unavailable(self):
        # pet itself is feedable, but pet is unavailable,
        # so cant feed pet should return False
        pet = Pet("Dragon-SolarSystem", feeding_status=FeedingStatus(-1))
        pair = PetMountPair(pet, pet_available=False, mount_available=True)

        result = pair.can_feed_pet()

        assert result is False

    @pytest.mark.parametrize(
        "unfeedable_pet_name",
        random.sample(PetData.unfeedable_pet_names, k=15)
        # Remark: The data is not entirely realistic, this includes some
        # mount-only animals too. However, they should also
        # return False when can_feed_pet is called.
    )
    def test_can_feed_pet_pet_is_unfeedable(self, unfeedable_pet_name: str):
        # pet itself is not feedable, so cant feed pet should return False
        pet = Pet(unfeedable_pet_name, feeding_status=FeedingStatus(-1))
        pair = PetMountPair(pet, pet_available=True, mount_available=False)

        result = pair.can_feed_pet()

        assert result is False


class TestZooBuilder:
    def test__init__(self):
        animal_name = "Wolf-Zombie"
        feeding_status = 5
        animal_name2 = "Deer-Base"
        feeding_status2 = 10

        pets_dict = {animal_name: feeding_status, animal_name2: feeding_status2}
        mounts_dict = {animal_name: True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets_dict, mounts=mounts_dict)

        builder = ZooBuilder(user)

        assert builder.pets == pets_dict
        assert builder.mounts == mounts_dict

    def test__repr__(self):
        animal_name = "Wolf-CottonCandyBlue"
        feeding_status = 5
        pets_dict = {animal_name: feeding_status}
        mounts_dict = {animal_name: True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets_dict, mounts=mounts_dict)

        builder_str = str(ZooBuilder(user))

        assert "'pets': " + str(pets_dict) in builder_str
        assert "'mounts': " + str(mounts_dict) in builder_str

    def test_build_empty_zoo(self):
        empty_zoo = {"pets": {}, "mounts": {}}
        user = HabiticaUser({"items": empty_zoo})

        zoo: Zoo = ZooBuilder(user).build()

        assert zoo == {}

    def test_build_raised_pet(self):
        animal_name = "BearCub-Shadow"
        feed_status = -1  # i.e. No pet, just mount
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets={animal_name: feed_status},
                                                        mounts={animal_name: True})

        zoo: Zoo = ZooBuilder(user).build()

        assert len(zoo) == 1

        result_pair = zoo[animal_name]
        assert result_pair.pet.pet_name == animal_name
        assert result_pair.pet.feeding_status == FeedingStatus(feed_status)
        assert result_pair.mount_available
        assert result_pair.pet_available is False

    def test_build_no_pet_yes_mount(self):
        animal_name = "Aether-Invisible"
        user: HabiticaUser = UserTestUtil.user_with_zoo(mounts={animal_name: True})

        zoo: Zoo = ZooBuilder(user).build()

        assert len(zoo) == 1
        result_pair = zoo[animal_name]
        assert result_pair.pet.pet_name == animal_name
        assert result_pair.mount_available
        assert result_pair.pet_available is False

    def test_build_yes_pet_no_mount(self):
        animal_name = "Dragon-Skeleton"
        feeding_status = 27
        user = UserTestUtil.user_with_zoo(pets={animal_name: feeding_status})

        zoo: Zoo = ZooBuilder(user).build()

        assert len(zoo) == 1
        result_pair = zoo[animal_name]
        assert result_pair.pet.pet_name == animal_name
        assert result_pair.pet.feeding_status == FeedingStatus(feeding_status)
        assert result_pair.mount_available is False
        assert result_pair.pet_available


class TestZooHelper:
    def test_filter_on_pet_mount_pair(self):
        pets = {"Owl-Golden": -1, "Ferret-Red": 27, "Phoenix-Base": 5, "Parrot-Base": 10}
        mounts = {"BearCub-Desert": True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets, mounts=mounts)
        zoo: Zoo = ZooBuilder(user).build()

        def predicate(pair: PetMountPair) -> bool:
            # arbitrary filter function
            return (pair.pet.pet_name == "Parrot-Base"
                    or pair.mount_available
                    or int(pair.pet.feeding_status) == 27)

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.filter_on_pet_mount_pairs(predicate=predicate)

        expected_pets = sorted(["Ferret-Red", "Parrot-Base", "BearCub-Desert"])
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)

    def test_get_feedable_zoo(self):
        pets = {
            "BearCub-Desert": 5,  # we have the mount, not hungry
            "Owl-Golden": -1,  # we don't have the pet
            "Ferret-Red": 27,  # can be fed
            "Phoenix-Base": 5,  # unfeedable pet
            "Parrot-Base": 10,  # can be fed
        }
        mounts: dict = {"BearCub-Desert": True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets, mounts=mounts)
        zoo: Zoo = ZooBuilder(user).build()

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.get_feedable_zoo()

        expected_pets = ["Ferret-Red", "Parrot-Base"]
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)

    def test_filter_on_pet_name(self):
        pets = {
            "BearCub-Desert": 5, "Owl-Golden": -1, "Ferret-Red": 27,
            "Phoenix-Base": 5, "Parrot-Base": 10
        }
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets)
        zoo: Zoo = ZooBuilder(user).build()

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.filter_on_pet_name(lambda name: name.endswith("-Base"))

        expected_pets = ["Phoenix-Base", "Parrot-Base"]
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)

    def test_filter_on_pet(self):
        pets: dict = {"BearCub-Desert": 9, "Owl-Golden": -1, "Ferret-Red": 27, "Phoenix-Base": 5,
                      "Parrot-Base": 10}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets)
        zoo: Zoo = ZooBuilder(user).build()

        def predicate(pet: Pet) -> bool:
            return int(pet.feeding_status) < 9

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.filter_on_pet(predicate=predicate)

        expected_pets = ["Owl-Golden", "Phoenix-Base"]
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)



"""
The module with CLI code that handles the `hopla buy` group command.
"""
import logging
import time

import click
import requests

from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.http import RequestHeaders, UrlBuilder
from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.cli.groupcmds.get_user import HabiticaUserRequest, HabiticaUser

log = logging.getLogger()


@click.group()
def buy():
    """
    GROUP to buy things.
    """


def buy_from_enchanted_armoire_once():
    """buy once from the enchanted armoire"""
    url = UrlBuilder(path_extension="/user/buy-armoire").url
    headers = RequestHeaders().get_default_request_headers()

    response = requests.post(url=url, headers=headers)
    buy_data = get_data_or_exit(response)

    # By default, we get way too much JSON info, so filter on "armoire".
    enchanted_armoire_award = JsonFormatter(buy_data["armoire"]).format_with_double_quotes()
    click.echo(enchanted_armoire_award)


def times_until_poor(gp_budget: float) -> int:
    """Returns how many times you can buy from enchanted_armoire given some budget.

    \f
    :param gp_budget:
    :return:
    """
    enchanted_armoire_price = 100
    times: float = gp_budget / enchanted_armoire_price
    return int(times)  # int will round down positive floats for us


@buy.command()
@click.option("--times", "-t", "requested_times",
              type=click.IntRange(min=0),
              metavar="[TIMES]",
              help="number of times to buy from the enchanted-armoire")
@click.option("--until-poor/--no-until-poor", "-u", "until_poor_flag",
              help="buy from enchanted-armoire until gp runs out",
              default=False, show_default=True)
def enchanted_armoire(requested_times: int, until_poor_flag: bool):
    """Buy from the enchanted armoire

    TIMES - the number of times to buy from the enchanted armoire. Must be at least 0.
    When TIMES is omitted, 1 is the default. When TIMES is larger than the budget allows,
    --until-poor will be used instead. This behavior is likely to change in the
    future: <https://github.com/melvio/hopla/issues/100>

    If no options are specified, you buy once.

    \b
    Examples
    ----
    # buy once
    $ hopla buy enchanted-armoire

    \b
    # buy 3 times
    $ hopla buy enchanted-armoire -t3
    $ hopla buy enchanted-armoire -t 3
    $ hopla buy enchanted-armoire --times 3
    $ hopla buy enchanted-armoire --times=3

    \b
    # buy until you cannot afford the enchanted-armoire anymore
    $ hopla buy enchanted-armoire --until-poor
    $ hopla buy enchanted-armoire -u
    """
    log.debug(f"hopla buy enchanted-armoire times={requested_times}, until_poor={until_poor_flag}")

    times: int = get_buy_times_within_budget(until_poor_flag=until_poor_flag,
                                             requested_times=requested_times)

    exceeds_throttle_threshold = exceeds_throttle_limit(times)
    throttle_seconds = 2.5
    if exceeds_throttle_threshold:
        click.echo(f"To prevent overheating Habitica, we'll "
                   f"buy once every {throttle_seconds} seconds")

    for _ in range(times):
        buy_from_enchanted_armoire_once()
        # throttle because we are going to call the API often
        if exceeds_throttle_threshold:
            time.sleep(throttle_seconds)


def exceeds_throttle_limit(times: int) -> bool:
    """Return True if we need Hopla to go easy on the Habitica API."""
    # Remark: When we need to throttle in multiple places, handle this globally, not
    #  here in `buy`.
    requests_throttle_limit = 25
    return times > requests_throttle_limit


def get_buy_times_within_budget(until_poor_flag: bool,
                                requested_times: int) -> int:
    """Return how often we can buy, given the requested amount and our budget."""
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    budget: float = user["stats"]["gp"]

    max_times = times_until_poor(budget)
    if until_poor_flag:
        requested_times = max_times
    elif requested_times is None:
        requested_times = 1

    times = min(max_times, requested_times)
    click.echo(f"The current budget is {int(budget)}gp.\n"
               f"Buying: {times} times.")
    return times



"""
The module with CLI code that handles the `hopla add` group command.
"""
import sys
import datetime
import logging
from typing import Dict, List, Optional

import click
import requests

from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.clickhelper import EnhancedDate
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder
from hopla.hoplalib.outputformatter import JsonFormatter

log = logging.getLogger()


@click.group()
def add():
    """GROUP for adding things to Habitica."""


DIFFICULTIES_SCORE_MAPPING = {"hard": "2", "medium": "1.5", "easy": "1", "trivial": "0.1"}


class HabiticaChecklist:
    """ Habitica Checklist """

    def __init__(self, *, checklist: Optional[list] = None):
        self.checklist = checklist or []

    def __repr__(self) -> object:
        return self.__class__.__name__ + f"(checklist={self.checklist})"

    def is_empty(self) -> bool:
        """Return true if the checklist is empty.

        >>> HabiticaChecklist(checklist=[]).is_empty()
        True
        >>> HabiticaChecklist(checklist=["item 1"]).is_empty()
        False

        """
        return self.checklist is None or len(self.checklist) == 0

    def to_json_list(self) -> List[Dict[str, str]]:
        """Turn a checklist into a list that we can pass to the habitica API"""
        return [{"text": line} for line in self.checklist]


class HabiticaTodo:
    """ Habitica To-Do"""

    def __init__(self, *,
                 todo_name: str,
                 difficulty: str,
                 due_date: Optional[datetime.datetime] = None,
                 checklist: Optional[HabiticaChecklist] = None):
        self.todo_name = todo_name
        self._type = "todo"
        self.difficulty = difficulty
        self.due_date = due_date
        self.checklist: Optional[HabiticaChecklist] = checklist or HabiticaChecklist()

    def difficulty_to_score(self) -> str:
        """Return score for the To-Do's difficulty.

        >>> hab_todo = HabiticaTodo(todo_name="my-task", difficulty="easy")
        >>> hab_todo.difficulty_to_score()
        '1'
        >>> hab_todo.difficulty = "trivial"
        >>> hab_todo.difficulty_to_score()
        '0.1'

        :raise KeyError in case of an invalid difficulty
        """
        return DIFFICULTIES_SCORE_MAPPING[self.difficulty]

    def due_date_to_date_str(self) -> str:
        """Turn the due_date into a valid ISO 8601 date format

        >>> hab_todo = HabiticaTodo(todo_name="my-task", difficulty="easy",\
                                    due_date=datetime.datetime(day=31, month=1, year=2069))
        >>> hab_todo.due_date_to_date_str()
        '2069-01-31'
        """
        habitica_date_format = "%Y-%m-%d"
        return self.due_date.strftime(habitica_date_format)

    def to_json_dict(self) -> dict:
        """Turn an instance of this class into a dict that we can pass to the habitica API.

        >>> hab_todo = HabiticaTodo(todo_name="my-task", difficulty="medium",\
                                    checklist=HabiticaChecklist(checklist=["myitem"]))
        >>> hab_todo.to_json_dict()
        {'text': 'my-task', 'type': 'todo', 'priority': '1.5', 'checklist': [{'text': 'myitem'}]}
        """
        todo_dict = {
            "text": self.todo_name,
            "type": self._type,
            "priority": self.difficulty_to_score(),
            "checklist": self.checklist.to_json_list()
        }

        if self.due_date:
            todo_dict["date"] = self.due_date_to_date_str()

        return todo_dict


class AddTodoRequest(HabiticaRequest):
    """An object that can perform an add To-Do request"""

    def __init__(self, habitica_todo: HabiticaTodo):
        self.url: str = UrlBuilder(path_extension="/tasks/user").url
        self.habitica_todo = habitica_todo

    def post_add_todo_request(self):
        """Perform the add To-Do request and return the data in case of success"""
        response = requests.post(
            url=self.url,
            headers=self.default_headers,
            json=self.habitica_todo.to_json_dict()
        )
        return get_data_or_exit(response)


# @see https://habitica.com/apidoc/#api-Task-CreateUserTasks
valid_difficulties = click.Choice(list(DIFFICULTIES_SCORE_MAPPING.keys()))


@add.command()
@click.option("--difficulty", type=valid_difficulties,
              default="easy",
              show_default=True,
              help="the priority of the To-Do")
@click.option("--due-date", "--deadline", type=EnhancedDate(),
              help="due date of the To-Do in format YYYY-MM-DD or DD-MM-YYYY."
                   "The special keywords 'today' and 'tomorrow' specify the "
                   "current day, or tomorrow.")
@click.option("--checklist", "checklist_file", type=click.File(),
              help="Every line in FILENAME will be an item of the checklist.")
@click.option("--editor/--no-editor", "checklist_editor", default=False,
              help="Open up an editor to create a checklist interactively",
              show_default=True)
@click.argument("todo_name", required=False)
def todo(difficulty: str,
         due_date: datetime.datetime,
         checklist_file,
         checklist_editor: bool,
         todo_name: str):
    """Add a To-Do.

    TODO_NAME the name of the To-Do. When omitted, hopla will prompt the user to input
    a TODO_NAME interactively.

    \b
    Examples:
    ---
    # Simplest way to create a To-Do
    $ hopla add todo "My todo"

    \b
    # If you don't specify a To-Do name, hopla will prompt you for one.
    $ hopla add todo
    Please provide a name for your todo: My todo


    \b
    # override the default 'easy' difficulty using either
    # --difficulty hard|medium|trivial
    $ hopla add --difficulty=hard todo 'This is a hard todo'

    \b
    # --due-date and --deadline work as synonyms.
    $ hopla add todo --deadline 2042-01-21 "My Todo"
    $ hopla add todo --due-date 2042-01-21 "My Todo"


    \b
    # You can use the 'today' or 'tomorrow' special keywords instead
    # of a date format.
    $ hopla add todo --due-date today "Feed pet"

    \b
    # You can use cool shell tricks to provide checklists on the fly
    # e.g. sort a list in a file:
    $ hopla add todo --checklist <(sort ./tasks.txt) "Sorted Checklist"

    \b
    # open up an editor and type in your new checklist on the go
    $ hopla add todo --editor "My Long Checklist"

    \b
    # open up a file, edit it, and send it as a checklist.
    # note that this keeps the underlying file unchanged
    $ hopla add todo --checklist tasks.txt --editor "My editable checklist"


    \b
    # create a hard To-Do with due date on the 22nd of october in 2077, and
    # use every line in the `my_todo.md` FILE as an item of your checklist.
    $ hopla add todo --difficulty hard --due-date=2077-10-21 \\
                   --checklist-file ./my_todo.md 'Text for my todo'


    [apidocs](https://habitica.com/apidoc/#api-Task-CreateUserTasks)

    \f
    :param checklist_editor:
    :param difficulty: str
    :param due_date:
    :param checklist_file:
    :param todo_name:
    :return:

    """
    log.debug(f"habitica add todo name={todo_name}"
              f"   difficulty={difficulty} , due_date={due_date}"
              f"   checklist ={checklist_file}, editor={checklist_editor}")

    habitica_todo = create_habitica_todo(checklist_editor=checklist_editor,
                                         checklist_file=checklist_file,
                                         difficulty=difficulty,
                                         due_date=due_date,
                                         todo_name=todo_name)

    add_todo_response_data = AddTodoRequest(habitica_todo=habitica_todo).post_add_todo_request()

    click.echo(JsonFormatter(add_todo_response_data).format_with_double_quotes())


def create_habitica_todo(*,
                         checklist_editor: bool,
                         checklist_file,
                         difficulty: str,
                         due_date: datetime.datetime,
                         todo_name: str) -> HabiticaTodo:
    """Create a HabiticaTodo object from the provided parameters"""
    habitica_checklist = get_checklist(checklist_file=checklist_file,
                                       checklist_editor=checklist_editor)
    if todo_name is None:
        todo_name = click.prompt("Please provide a name for your todo")
    habitica_todo = HabiticaTodo(todo_name=todo_name, difficulty=difficulty,
                                 due_date=due_date, checklist=habitica_checklist)
    return habitica_todo


def get_checklist(checklist_file, checklist_editor: bool) -> HabiticaChecklist:
    """Get a habitica checklist from the user

     Returns an empty HabiticaChecklist if user did not want a checklist

     >>> get_checklist(None, False)
     HabiticaChecklist(checklist=[])
    """
    if checklist_editor:
        return get_checklist_with_editor(checklist_file)
    if checklist_file:
        return HabiticaChecklist(checklist=checklist_file.readlines())
    return HabiticaChecklist(checklist=[])


def get_checklist_with_editor(checklist_file) -> HabiticaChecklist:
    """
    Get the HabiticaUser for an user that requested the editor.
    """
    comment = "# Every line below this comment represents an item on your checklist\n"
    if checklist_file:
        text = comment + "".join(checklist_file.readlines())
    else:
        text = comment
    message = click.edit(text=text, extension=".md")
    if message is not None:
        try:
            checklist_str_with_comment: List[str] = message.split(comment, maxsplit=1)
        except IndexError as ex:
            sys.exit(f"could not find the checklist below the specified comment: {ex}")
        checklist_str = checklist_str_with_comment[1]
        return HabiticaChecklist(checklist=checklist_str.split("\n"))

    sys.exit("editor exited")



"""
A helper module for Pet logic.
"""
from dataclasses import dataclass
from typing import Callable, Dict, NoReturn, Optional

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.errors import PrintableException
from hopla.hoplalib.zoo.fooddata import FoodData
from hopla.hoplalib.zoo.foodmodels import FeedingStatus, InvalidFeedingStatus
from hopla.hoplalib.zoo.petdata import PetData


class InvalidPet(PrintableException):
    """Exception raised when a pet is invalid."""

    def __init__(self, msg: str, *, pet: Optional["Pet"] = None):
        super().__init__(msg)
        self.pet = pet


class Pet:
    """A habitica pet"""

    # pylint: disable=too-many-public-methods
    #################################################################
    # In a way, pylint is right. This class does have a lot of functions.
    # Most of these are oneliners, however we can also use polymorphism.
    # That would make it easier to find functions in this class.
    # We can use a PetFactory that creates QuestPets, MagicPets etc.
    # However, that is really overkill for now and not urgent enough.
    # Therefore, too-many-public-methods is disabled for this class.
    #################################################################

    def __init__(self, pet_name: str, *,
                 feeding_status: FeedingStatus = FeedingStatus(5)):
        if pet_name not in PetData.pet_names:
            raise InvalidPet(f"{pet_name=} is not recognized by hopla.\n"
                             "Potential causes: \n"
                             "* did you spell it correctly?\n"
                             "* Is your pet relatively new? If so, please raise\n"
                             f" an issue at {GlobalConstants.NEW_ISSUE_URL}")

        self.pet_name = pet_name
        self.feeding_status = feeding_status

        if pet_name in PetData.rare_pet_names:
            self.__potion = None
        else:
            # This logic might break, but seems to be solid for past few years due to
            # stabling naming convention by the Habitica API developers.
            _, self.__potion = self.pet_name.split("-")

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.pet_name=}, {self.feeding_status=})"

    @property
    def hatching_potion(self):
        """The hatching potion used to hatch the egg this pet came from."""
        return self.__potion

    def is_feedable(self) -> bool:
        """Return True if a pet cannot be fed at all."""
        return self.pet_name not in PetData.unfeedable_pet_names

    def has_just_1_favorite_food(self) -> bool:
        """Return True if this pet likes only 1 type of food."""
        return self.pet_name in PetData.only_1favorite_food_pet_names

    def likes_all_food(self) -> bool:
        """Return True if this pet prefers all food."""
        return self.pet_name in PetData.magic_potion_pet_names

    def feeding_status_explanation(self) -> str:
        """Explain the feeding status of a pet."""
        if self.is_feedable() is False:
            return f"{self.pet_name=} can't be fed because it is special."
        if int(self.feeding_status) == -1:
            return f"You can't feed {self.pet_name=} you only have the mount"

        if int(self.feeding_status) == 5:
            # remark: if we really want to know more, we can query the
            #         API to check if we have the self.pet_name in the
            #         Users: jq .data.items.mounts
            return (f"Cannot determine if {self.pet_name=} can be fed. \n"
                    "You either have:\n"
                    "1. Both the pet and mount. In this case, you"
                    "   cannot feed the pet \n"
                    "2. A pet that hasn't been fed but you don't have \n"
                    "   the mount. In this case, you can feed your pet")

        if int(self.feeding_status) < 50:
            return f"{self.pet_name=} can be fed"

        raise InvalidPet(f"Did not expect {self.feeding_status=}. \n"
                         f"Looks like we failed to validate the Pet properly.",
                         pet=self)  # pragma: no cover

    def favorite_food(self, *,
                      default_value_for_unfeedable: str = "Unfeedable",
                      default_value_for_all_favorite_food: str = "Any") -> str:
        """Return the favorite food of this pet."""
        if self.is_feedable() is False:
            return default_value_for_unfeedable

        if self.pet_name in PetData.magic_potion_pet_names:
            return default_value_for_all_favorite_food

        if self.has_just_1_favorite_food():
            return (FoodData.hatching_potion_favorite_food_mapping
                    .get(self.hatching_potion))

        raise InvalidPet(f"Could not find the feeding habits of this {self.pet_name=}",
                         pet=self)  # pragma: no cover

    def is_favorite_food(self, food_name: str) -> bool:
        """Return true if 'food_name' is this Pets favorite food."""
        if self.is_feedable() is False:
            return False
        if self.likes_all_food():
            return True
        if self.has_just_1_favorite_food():
            return self.favorite_food() == food_name

        raise InvalidPet(f"Unable to determine if {food_name=} is {self.pet_name=}'s favorite.",
                         pet=self)  # pragma: no cover

    def required_food_items_until_mount(self, food_name: str) -> int:
        """
        Return the number of times food_name needs to be fed
        for this pet to turn into a mount
        """
        is_favorite: bool = self.is_favorite_food(food_name)
        return self.feeding_status.required_food_items_to_become_mount(is_favorite)

    def is_generation1_pet(self) -> bool:
        """Return True if this pet is from the generation 1 pet"""
        return self.pet_name in PetData.generation1_pet_names

    def is_quest_pet(self) -> bool:
        """
        Return True if this pet is a quest pet. This doesn't include
        special pets such as world event related pets.
        """
        return self.pet_name in PetData.quest_pet_names

    def is_magic_hatching_pet(self) -> bool:
        """Return True if this pet is hatched from a magic potion."""
        return self.pet_name in PetData.magic_potion_pet_names

    def is_from_drop_hatching_potions(self) -> bool:
        """
        Return True if the pet was hatched from one of the 'ordinary'
        potions. (Such as: Base, Desert, ...).
        """
        return self.hatching_potion in FoodData.drop_hatching_potions


class PetMountPair:
    """A pair of a pet and its corresponding mount information."""

    def __init__(self, pet: Pet, *,
                 pet_available: bool,
                 mount_available: bool):
        """Create a Pet-Mount pair"""
        if pet.feeding_status is None:
            msg = f"PetMountPair requires that {pet=} has a feeding status"
            raise InvalidFeedingStatus(msg)

        self.pet = pet
        self.pet_available = pet_available
        self.mount_available = mount_available

    def __repr__(self):
        return self.__class__.__name__ + f"({self.__dict__})"

    def can_feed_pet(self) -> True:
        """Return true if the pet itself is feedable and there is no mount yet."""
        return (
                self.mount_available is False
                and self.pet_available
                and self.pet.is_feedable()
                and int(self.pet.feeding_status) != -1
        )


Zoo = Dict[str, PetMountPair]
"""
Zoo is dictionary with key pet name keys for O(1) access to the
PetMountPair.
"""


@dataclass(frozen=True)
class ZooHelper:
    """Class with helper functions for a Zoo."""
    zoo: Zoo

    def filter_on_pet_mount_pairs(self, predicate: Callable[[PetMountPair], bool]) -> Zoo:
        """Filter the zoo on the pair. This does not change the underlying zoo.

        :param predicate: Include the PetMountPair if the predicate returns True, else
                          omit the pair.
        :return: A filtered Zoo
        """
        return {
            pet_name: pair for (pet_name, pair) in self.zoo.items() if predicate(pair)
        }

    def get_feedable_zoo(self) -> Zoo:
        """Helper function to get only the pets that can be fed in this Zoo."""
        return self.filter_on_pet_mount_pairs(PetMountPair.can_feed_pet)

    def filter_on_pet_name(self, predicate: Callable[[str], bool]) -> Zoo:
        """Filter the zoo on the pet name. This does not change the underlying zoo.

        :param predicate: Include the PetMountPair if the predicate returns True, else
                          omit the pair.
        :return: A filtered zoo.
        """
        return {
            pet_name: pair for (pet_name, pair) in self.zoo.items() if predicate(pet_name)
        }

    def filter_on_pet(self, predicate: Callable[[Pet], bool]) -> Zoo:
        """Filter the zoo on the pet object. This does not change the underlying zoo.

        :param predicate: Include the PetMountPair if the predicate returns True, else
                          omit the pair.
        :return: A filtered zoo.
        """
        return {
            pet_name: pair for (pet_name, pair) in self.zoo.items() if predicate(pair.pet)
        }


class ZooBuilder:
    """
    Class that creates a Zoo from a HabiticaUser using the
    builder design pattern.
    """

    def __init__(self, user: HabiticaUser):
        self.pets: dict = user.get_pets()
        self.mounts: dict = user.get_mounts()
        self.__zoo: Zoo = {}  # empty until build() is called

    def __repr__(self):
        return self.__class__.__name__ + f"({self.__dict__})"

    def build(self) -> Zoo:
        """Create the Zoo. Return the Zoo in case of success"""
        self.__build()
        return self.__zoo

    def __build(self) -> NoReturn:
        self.__add_pets_to_zoo()
        self.__add_remaining_mounts_to_zoo()

    def __add_pets_to_zoo(self) -> NoReturn:
        for pet_name in self.pets:
            feed_status = self.pets[pet_name]
            if self.mounts.get(pet_name) is None:
                mount_available = False
            else:
                mount_available = True
                del self.mounts[pet_name]  # found the mount!

            pet_available = feed_status != -1
            pet = Pet(pet_name, feeding_status=FeedingStatus(feed_status))
            pair = PetMountPair(pet,
                                pet_available=pet_available,
                                mount_available=mount_available)

            self.__zoo[pet_name] = pair

    def __add_remaining_mounts_to_zoo(self) -> NoReturn:
        for mount_name in self.mounts:
            # Watch out: we are abusing the Pet object for mount only animals here
            pet = Pet(mount_name)

            pair = PetMountPair(pet,
                                pet_available=False,
                                mount_available=True)
            self.__zoo[mount_name] = pair



"""
Module responsible for setting up initialization of the hopla entry command.
"""
import logging
import sys

import click

from hopla.cli.authenticate import authenticate
from hopla.cli.complete import complete
from hopla.cli.config import config
from hopla.cli.feed import feed
from hopla.cli.feed_all import feed_all
from hopla.cli.get_group import get_group
from hopla.cli.get_user.auth import auth
from hopla.cli.get_user.info import info
from hopla.cli.get_user.inventory import inventory
from hopla.cli.get_user.stats import stats
from hopla.cli.groupcmds.add import add
from hopla.cli.groupcmds.api import api
from hopla.cli.groupcmds.buy import buy
from hopla.cli.groupcmds.get_user import get_user
from hopla.cli.groupcmds.set import set  # pylint: disable=redefined-builtin
from hopla.cli.request import request
from hopla.cli.support_development import support_development
from hopla.cli.version import version
from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.configuration import ConfigInitializer, ConfigurationFileParser


def setup_logging() -> logging.Logger:
    """Setup python logging for the entire hopla project"""

    parsed_loglevel = ConfigurationFileParser().get_full_config_name("cmd_all.loglevel",
                                                                     fallback="info")
    loglevel_mapping = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR
    }

    # https://docs.python.org/3.8/howto/logging.html#logging-basic-tutorial
    logging.basicConfig(
        format="[%(levelname)s][%(filename)s|%(asctime)s] %(message)s",
        level=loglevel_mapping[parsed_loglevel],
        datefmt="%Y-%m-%dT%H:%M:%S"
    )
    return logging.getLogger(__name__)


log = setup_logging()

HOPLA_CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],  # add -h
    auto_envvar_prefix=GlobalConstants.APPLICATION_NAME
)


@click.group(context_settings=HOPLA_CONTEXT_SETTINGS)
@click.version_option()
def hopla():
    """hopla - a command line interface (CLI) to interact with habitica.com"""


def entry_cmd():
    """The entry function for the hopla command"""
    had_to_create_new_config_file = ConfigInitializer().initialize_before_running_cmds()
    if had_to_create_new_config_file:
        click.echo("Thank you for trying out hopla in its early release")
        click.echo(
            "Bug reports, pull requests, and feature requests are welcomed over at:  ")
        click.echo("  <https://github.com/melvio/hopla>")
    log.debug(f"start application with arguments: {sys.argv}")

    organize_cli()
    hopla()


def organize_cli():
    """Attach the subgroups and subcommands to the top hopla group command"""
    # subgroups
    hopla.add_command(add)
    hopla.add_command(api)
    hopla.add_command(set)
    hopla.add_command(buy)
    hopla.add_command(get_user)
    # get-user subgroup
    get_user.add_command(inventory)
    get_user.add_command(stats)
    get_user.add_command(info)
    get_user.add_command(auth)
    # subcommands
    hopla.add_command(config)
    hopla.add_command(complete)
    hopla.add_command(version)
    hopla.add_command(authenticate)
    hopla.add_command(feed)
    hopla.add_command(feed_all)
    hopla.add_command(support_development)
    hopla.add_command(request)
    hopla.add_command(get_group)



"""
Module with commonly used python-click functionality.
"""
import logging
import datetime
from typing import List, Optional, Any

import click

log = logging.getLogger()




def case_insensitive_aliases(argument: str, choices: List[str], *,
                             raise_if_not_found=False):
    """Takes a potentially in-properly cased argument and returns
       the properly cased version"""
    # This function works but it very ugly:
    # @click.command(context_settings=dict(
    #       token_normalize_func= lambda x: case_insensitive_aliases(x, valid_food_names))
    # )

    for choice in choices:
        if argument.lower() == choice.lower():
            return choice

    if raise_if_not_found:
        raise ValueError(f"Could not match argument={argument} with choices={choices}")
    return argument


class EnhancedDate(click.DateTime):
    """EnhancedDate adds 'today' and 'tomorrow' keywords to the click.DateTime type."""
    name = "enhanceddate"

    def __init__(self):
        super().__init__(formats=["%Y-%m-%d", "%d-%m-%Y"])

        # This will add 'today' and 'tomorrow to the help message
        self.formats += ["today", "tomorrow"]

    def convert(self, value: Any, param: Optional["Parameter"], ctx: Optional["Context"]) -> Any:
        """Convert a value provided by the user into a datetime object."""
        if value == "today":
            return datetime.date.today()
        if value == "tomorrow":
            return datetime.date.today() + datetime.timedelta(days=1)

        return super().convert(value, param, ctx)



import setuptools

setuptools.setup()

# [quote](https://setuptools.readthedocs.io/en/latest/setuptools.html)
# "As PEP 517 is new, support is not universal, and frontends that do
# support it may still have bugs. For compatibility, you may want to
# put a setup.py file containing only a setuptools.setup() invocation."




#!/usr/bin/env python3
from unittest.mock import MagicMock, patch

import pytest

from hopla.hoplalib.zoo.petcontroller import FeedPostRequester


class TestFeedPostRequester:
    def test_path_is_ok(self):
        pet = "Wolf-Golden"
        food = "Honey"

        feed_requester = FeedPostRequester(pet_name=pet,
                                           food_name=food)

        assert feed_requester.path == f"/user/feed/{pet}/{food}"

    def test_url_is_okay(self):
        pet = "Wolf-Shimmer"
        food = "Meat"

        feed_requester = FeedPostRequester(pet_name=pet,
                                           food_name=food)

        result_url = feed_requester.feed_pet_food_url

        expected_url = f"https://habitica.com/api/v3/user/feed/{pet}/{food}"
        assert result_url == expected_url

    @pytest.mark.parametrize("times,expected_times", [(2, 2), (None, 1)])
    @patch("hopla.hoplalib.zoo.petcontroller.HabiticaRequest.default_headers")
    @patch("hopla.hoplalib.zoo.petcontroller.requests.post")
    def test_post_feed_request(self, mock_post_request: MagicMock,
                               mock_headers: MagicMock,
                               times: int,
                               expected_times: int):
        pet_name = "Egg-CottonCandyPink"
        food_name = "CottonCandyPink"

        feed_requester = FeedPostRequester(
            pet_name=pet_name, food_name=food_name, food_amount=times
        )

        _ = feed_requester.post_feed_request()

        expected_url = f"https://habitica.com/api/v3/user/feed/{pet_name}/{food_name}"
        mock_post_request.assert_called_with(
            url=expected_url,
            headers=mock_headers,
            params={"amount": expected_times}
        )



"""
The module with CLI code that handles the `hopla get` group command.
"""
import logging
from dataclasses import dataclass
from typing import Dict

import click
import requests

from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder

log = logging.getLogger()


@dataclass(frozen=True)
class HabiticaUser:
    """
    Class representing a user model.

    The user_dict is assumed to be returned from a 200 ok Response as (using
    Response.json()) when calling the /user endpoint and getting .data
    """
    user_dict: dict

    def __getitem__(self, key):
        return self.user_dict.__getitem__(key)

    def get_stats(self) -> dict:
        """Index the user_dict for 'stats' and return the result"""
        return self["stats"]

    def get_inventory(self) -> dict:
        """Index the user_dict for 'items' and return the result"""
        return self["items"]

    def get_pets(self) -> dict:
        """Return the pets of a user."""
        return self.get_inventory()["pets"]

    def get_mounts(self) -> dict:
        """Return the mounts of a user."""
        return self.get_inventory()["mounts"]

    def get_food(self) -> Dict[str, int]:
        """Return the food that the user has"""
        return self.get_inventory()["food"]

    def get_auth(self) -> dict:
        """Index the user_dict for 'auth' and return the result"""
        return self["auth"]

    def get_gems(self) -> int:
        """Get the number of gems of a user.

        gems are stored in the 'balance' field. 1 'balance' equals 4 gems
        [see](https://habitica.fandom.com/wiki/Gems#Information_for_Developers)
        """
        balance = self["balance"]
        return int(balance * 4)


class HabiticaUserRequest(HabiticaRequest):
    """Class that requests a user model from the Habitica API"""

    def __init__(self):
        self.url = UrlBuilder(path_extension="/user").url

    def request_user(self) -> requests.Response:
        """Perform the user get request and return the response"""
        return requests.get(url=self.url, headers=self.default_headers)

    def request_user_data_or_exit(self) -> HabiticaUser:
        """
        Function that request the user from habitica and returns
        a HabiticaUser if the request was successful. Else exits.
        """
        user_response: requests.Response = self.request_user()
        user_data: dict = get_data_or_exit(user_response)
        return HabiticaUser(user_dict=user_data)


pass_user = click.make_pass_decorator(HabiticaUser)


@click.group()
@click.pass_context
def get_user(ctx: click.Context) -> HabiticaUser:
    """
    GROUP for getting user information from Habitica.
    """
    log.debug("hopla get-user")
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    ctx.obj = user
    return user
